#! /usr/bin/env bash
# Copyright (c) 2011-2013, ARM Limited
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of ARM nor the names of its contributors may be used
#       to endorse or promote products derived from this software without
#       specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

set -e
set -x
set -u
set -o pipefail

umask 022

exec < /dev/null

script_path=`cd $(dirname $0) && pwd -P`
. $script_path/build-common_freebsd9.sh

# This file contains the sequence of commands used to build the ARM EABI toolchain.
usage ()
{
    echo "Usage:" >&2
    echo "      $0 [--debug]" >&2
    exit 1
}
if [ $# -gt 1 ] ; then
    usage
fi
DEBUG_BUILD_OPTIONS=no
MULTILIB_LIST="--with-multilib-list=armv6-m,armv7-m,armv7e-m,armv7-r"
for ac_arg; do
    case $ac_arg in
        --debug)
            DEBUG_BUILD_OPTIONS=" -O0 -g "
            ;;
        *)
            usage
            ;;
    esac
done

ENV_CFLAGS=" -I/usr/lib/include -O2 "
ENV_CPPFLAGS=" -I/usr/lib/include "
ENV_LDFLAGS=" -L/usr/lib
              -L/usr/local/lib "

GCC_CONFIG_OPTS=" --build=$BUILD --host=$HOST_NATIVE
                  --with-gmp=/usr/local
                  --with-mpfr=/usr/local
                  --with-mpc=/usr/local
                  --with-isl=$BUILDDIR_NATIVE/host-libs/usr
                  --with-cloog=/usr/local
                  --with-libelf=/usr/local "

BINUTILS_CONFIG_OPTS=" --build=$BUILD --host=$HOST_NATIVE "

NEWLIB_CONFIG_OPTS=" --build=$BUILD --host=$HOST_NATIVE "

GDB_CONFIG_OPTS=" --build=$BUILD --host=$HOST_NATIVE
                  --with-libexpat-prefix=/usr/local "
mkdir -p $BUILDDIR_NATIVE
rm -rf $INSTALLDIR_NATIVE && mkdir -p $INSTALLDIR_NATIVE
rm -rf $PACKAGEDIR && mkdir -p $PACKAGEDIR

cd $SRCDIR

echo Task [III-0] /$HOST_NATIVE/binutils/
rm -rf $BUILDDIR_NATIVE/binutils && mkdir -p $BUILDDIR_NATIVE/binutils
pushd $BUILDDIR_NATIVE/binutils
saveenv
saveenvvar CFLAGS "$ENV_CFLAGS"
saveenvvar CPPFLAGS "$ENV_CPPFLAGS"
saveenvvar LDFLAGS "$ENV_LDFLAGS"
$SRCDIR/$BINUTILS/configure                         \
    ${BINUTILS_CONFIG_OPTS}                         \
    --target=$TARGET                                \
    --prefix=$INSTALLDIR_NATIVE                     \
    --infodir=$INSTALLDIR_NATIVE_DOC/info           \
    --mandir=$INSTALLDIR_NATIVE_DOC/man             \
    --htmldir=$INSTALLDIR_NATIVE_DOC/html           \
    --disable-nls                                   \
    --enable-plugins                                \
    --with-sysroot=$INSTALLDIR_NATIVE/arm-none-eabi \
    "--with-pkgversion=$PKGVERSION"

if [ "x$DEBUG_BUILD_OPTIONS" != "xno" ] ; then
    gmake CFLAGS="-I/usr/include $DEBUG_BUILD_OPTIONS" -j$JOBS
else
    gmake -j$JOBS
fi

gmake install install-html

copy_dir $INSTALLDIR_NATIVE $BUILDDIR_NATIVE/target-libs
restoreenv
popd

pushd $INSTALLDIR_NATIVE
rm -rf ./lib
popd

echo Task [III-1] /$HOST_NATIVE/gcc-first/
rm -rf $BUILDDIR_NATIVE/gcc-first && mkdir -p $BUILDDIR_NATIVE/gcc-first
pushd $BUILDDIR_NATIVE/gcc-first
$SRCDIR/$GCC/configure --target=$TARGET             \
    --prefix=$INSTALLDIR_NATIVE                     \
    --libexecdir=$INSTALLDIR_NATIVE/lib             \
    --infodir=$INSTALLDIR_NATIVE_DOC/info           \
    --mandir=$INSTALLDIR_NATIVE_DOC/man             \
    --htmldir=$INSTALLDIR_NATIVE_DOC/html           \
    --enable-languages=c                            \
    --disable-decimal-float                         \
    --disable-libffi                                \
    --disable-libgomp                               \
    --disable-libmudflap                            \
    --disable-libquadmath                           \
    --disable-libssp                                \
    --disable-libstdcxx-pch                         \
    --disable-nls                                   \
    --disable-shared                                \
    --disable-threads                               \
    --disable-tls                                   \
    --with-newlib                                   \
    --without-headers                               \
    --with-gnu-as                                   \
    --with-gnu-ld                                   \
    --with-python-dir=share/gcc-arm-none-eabi       \
    --with-sysroot=$INSTALLDIR_NATIVE/arm-none-eabi \
    ${GCC_CONFIG_OPTS}                              \
    "${GCC_CONFIG_OPTS_LCPP}"                       \
    "--with-pkgversion=$PKGVERSION"                 \
    ${MULTILIB_LIST}

gmake -j$JOBS all-gcc

gmake install-gcc

popd

pushd $INSTALLDIR_NATIVE
rm -rf bin/arm-none-eabi-gccbug
rm -rf ./lib/libiberty.a
rmdir include
popd

echo Task [III-2] /$HOST_NATIVE/newlib/
saveenv
prepend_path PATH $INSTALLDIR_NATIVE/bin
saveenvvar CFLAGS_FOR_TARGET '-g -O2 -fno-unroll-loops -ffunction-sections -fdata-sections'
rm -rf $BUILDDIR_NATIVE/newlib && mkdir -p $BUILDDIR_NATIVE/newlib
pushd $BUILDDIR_NATIVE/newlib

$SRCDIR/$NEWLIB/configure                 \
    $NEWLIB_CONFIG_OPTS                   \
    --target=$TARGET                      \
    --prefix=$INSTALLDIR_NATIVE           \
    --infodir=$INSTALLDIR_NATIVE_DOC/info \
    --mandir=$INSTALLDIR_NATIVE_DOC/man   \
    --htmldir=$INSTALLDIR_NATIVE_DOC/html \
    --enable-newlib-io-long-long          \
    --enable-newlib-register-fini         \
    --disable-newlib-supplied-syscalls    \
    --disable-nls

gmake -j$JOBS

gmake install
gmake html
mkdir -p $INSTALLDIR_NATIVE_DOC/html
copy_dir $BUILDDIR_NATIVE/newlib/arm-none-eabi/newlib/libc/libc.html $INSTALLDIR_NATIVE_DOC/html/libc
copy_dir $BUILDDIR_NATIVE/newlib/arm-none-eabi/newlib/libm/libm.html $INSTALLDIR_NATIVE_DOC/html/libm

popd
restoreenv

echo Task [III-3] /$HOST_NATIVE/newlib-nano/
saveenv
prepend_path PATH $INSTALLDIR_NATIVE/bin
saveenvvar CFLAGS_FOR_TARGET '-g -Os -ffunction-sections -fdata-sections'
rm -rf $BUILDDIR_NATIVE/newlib-nano && mkdir -p $BUILDDIR_NATIVE/newlib-nano
pushd $BUILDDIR_NATIVE/newlib-nano

$SRCDIR/$NEWLIB_NANO/configure            \
    $NEWLIB_CONFIG_OPTS                   \
    --target=$TARGET                      \
    --prefix=$BUILDDIR_NATIVE/target-libs \
    --disable-newlib-supplied-syscalls    \
    --enable-newlib-reent-small           \
    --disable-newlib-fvwrite-in-streamio  \
    --disable-newlib-fseek-optimization   \
    --disable-newlib-wide-orient          \
    --enable-newlib-nano-malloc           \
    --disable-newlib-unbuf-stream-opt     \
    --enable-lite-exit                    \
    --enable-newlib-global-atexit         \
    --disable-nls

gmake -j$JOBS
gmake install

popd
restoreenv

echo Task [III-4] /$HOST_NATIVE/gcc-final/
rm -f $INSTALLDIR_NATIVE/arm-none-eabi/usr
ln -s . $INSTALLDIR_NATIVE/arm-none-eabi/usr

rm -rf $BUILDDIR_NATIVE/gcc-final && mkdir -p $BUILDDIR_NATIVE/gcc-final
pushd $BUILDDIR_NATIVE/gcc-final

$SRCDIR/$GCC/configure --target=$TARGET             \
    --prefix=$INSTALLDIR_NATIVE                     \
    --libexecdir=$INSTALLDIR_NATIVE/lib             \
    --infodir=$INSTALLDIR_NATIVE_DOC/info           \
    --mandir=$INSTALLDIR_NATIVE_DOC/man             \
    --htmldir=$INSTALLDIR_NATIVE_DOC/html           \
    --enable-languages=c,c++                        \
    --enable-plugins                                \
    --disable-decimal-float                         \
    --disable-libffi                                \
    --disable-libgomp                               \
    --disable-libmudflap                            \
    --disable-libquadmath                           \
    --disable-libssp                                \
    --disable-libstdcxx-pch                         \
    --disable-nls                                   \
    --disable-shared                                \
    --disable-threads                               \
    --disable-tls                                   \
    --with-gnu-as                                   \
    --with-gnu-ld                                   \
    --with-newlib                                   \
    --with-headers=yes                              \
    --with-python-dir=share/gcc-arm-none-eabi       \
    --with-sysroot=$INSTALLDIR_NATIVE/arm-none-eabi \
    $GCC_CONFIG_OPTS                                \
    "${GCC_CONFIG_OPTS_LCPP}"                       \
    "--with-pkgversion=$PKGVERSION"                 \
    ${MULTILIB_LIST}

# Passing USE_TM_CLONE_REGISTRY=0 via INHIBIT_LIBC_CFLAGS to disable
# transactional memory related code in crtbegin.o.
# This is a workaround. Better approach is have a t-* to set this flag via
# CRTSTUFF_T_CFLAGS
if [ "x$DEBUG_BUILD_OPTIONS" != "xno" ]; then
  gmake -j$JOBS CFLAGS="$DEBUG_BUILD_OPTIONS" \
	       INHIBIT_LIBC_CFLAGS="-DUSE_TM_CLONE_REGISTRY=0"
else
  gmake -j$JOBS INHIBIT_LIBC_CFLAGS="-DUSE_TM_CLONE_REGISTRY=0"
fi

gmake install install-html

pushd $INSTALLDIR_NATIVE
rm -rf bin/arm-none-eabi-gccbug
LIBIBERTY_LIBRARIES=`find $INSTALLDIR_NATIVE/arm-none-eabi/lib -name libiberty.a`
for libiberty_lib in $LIBIBERTY_LIBRARIES ; do
    rm -rf $libiberty_lib
done
rm -rf ./lib/libiberty.a
rmdir include
popd

rm -f $INSTALLDIR_NATIVE/arm-none-eabi/usr
popd


echo Task [III-4.1] /$HOST_NATIVE/gcc-plugins
#build and install GCC plugins
if [ -d $SRCDIR/$GCC_PLUGINS/ ] && [ "x$build_gcc_plugin" == "xyes" ]; then
plugin_dir=$($INSTALLDIR_NATIVE/bin/arm-none-eabi-gcc -print-file-name=plugin)
# search for all directories not starting with .
plugin_src_dirs=$(find $SRCDIR/$GCC_PLUGINS/ -mindepth 1 -maxdepth 1 -type d -name '[^\.]*')
for d in $plugin_src_dirs; do
    plugin_name=$(basename $d)
    src_files=$(find $d -name \*.c -or -name \*.cc)
    g++ -fPIC -fno-rtti -O2 -shared -I /usr/local/include/ -I $plugin_dir/include $src_files -o $plugin_dir/$plugin_name.so
done
fi

echo Task [III-5] /$HOST_NATIVE/gcc-size-libstdcxx/
rm -f $BUILDDIR_NATIVE/target-libs/arm-none-eabi/usr
ln -s . $BUILDDIR_NATIVE/target-libs/arm-none-eabi/usr

rm -rf $BUILDDIR_NATIVE/gcc-size-libstdcxx && mkdir -p $BUILDDIR_NATIVE/gcc-size-libstdcxx
pushd $BUILDDIR_NATIVE/gcc-size-libstdcxx

$SRCDIR/$GCC/configure --target=$TARGET                       \
    --prefix=$BUILDDIR_NATIVE/target-libs                     \
    --enable-languages=c,c++                                  \
    --disable-decimal-float                                   \
    --disable-libffi                                          \
    --disable-libgomp                                         \
    --disable-libmudflap                                      \
    --disable-libquadmath                                     \
    --disable-libssp                                          \
    --disable-libstdcxx-pch                                   \
    --disable-nls                                             \
    --disable-shared                                          \
    --disable-threads                                         \
    --disable-tls                                             \
    --with-gnu-as                                             \
    --with-gnu-ld                                             \
    --with-newlib                                             \
    --with-headers=yes                                        \
    --with-python-dir=share/gcc-arm-none-eabi                 \
    --with-sysroot=$BUILDDIR_NATIVE/target-libs/arm-none-eabi \
    $GCC_CONFIG_OPTS                                          \
    "${GCC_CONFIG_OPTS_LCPP}"                                 \
    "--with-pkgversion=$PKGVERSION"                           \
    ${MULTILIB_LIST}

gmake -j$JOBS CXXFLAGS_FOR_TARGET="-g -Os -ffunction-sections -fdata-sections -fno-exceptions"
gmake install

copy_multi_libs src_prefix="$BUILDDIR_NATIVE/target-libs/arm-none-eabi/lib" \
                dst_prefix="$INSTALLDIR_NATIVE/arm-none-eabi/lib"           \
                target_gcc="$BUILDDIR_NATIVE/target-libs/bin/arm-none-eabi-gcc"
popd

echo Task [III-6] /$HOST_NATIVE/gdb/
rm -rf $BUILDDIR_NATIVE/gdb && mkdir -p $BUILDDIR_NATIVE/gdb
pushd $BUILDDIR_NATIVE/gdb
saveenv
saveenvvar CFLAGS "$ENV_CFLAGS"
saveenvvar CPPFLAGS "$ENV_CPPFLAGS"
saveenvvar LDFLAGS "$ENV_LDFLAGS"
$SRCDIR/$GDB/configure                                                              \
    --target=$TARGET                                                                \
    --prefix=$INSTALLDIR_NATIVE                                                     \
    --infodir=$INSTALLDIR_NATIVE_DOC/info                                           \
    --mandir=$INSTALLDIR_NATIVE_DOC/man                                             \
    --htmldir=$INSTALLDIR_NATIVE_DOC/html                                           \
    --disable-nls                                                                   \
    --disable-sim                                                                   \
    --with-libexpat                                                                 \
    --with-python=no                                                                \
    --with-system-gdbinit=$INSTALLDIR_NATIVE/$HOST_NATIVE/arm-none-eabi/lib/gdbinit \
    $GDB_CONFIG_OPTS                                                                \
    '--with-gdb-datadir='\''${prefix}'\''/arm-none-eabi/share/gdb'                  \
    "--with-pkgversion=$PKGVERSION"

if [ "x$DEBUG_BUILD_OPTIONS" != "xno" ] ; then
    gmake CFLAGS="-I/usr/include $DEBUG_BUILD_OPTIONS" -j$JOBS
else
    gmake -j$JOBS
fi

gmake install install-html 

restoreenv
popd

echo TASK [III-7] /$HOST_NATIVE/build-manual
rm -rf $BUILDDIR_NATIVE/build-manual && mkdir -p $BUILDDIR_NATIVE/build-manual
pushd $BUILDDIR_NATIVE/build-manual
cp -r $SRCDIR/$BUILD_MANUAL/* .
echo "@set VERSION_PACKAGE ($PKGVERSION)" > version.texi
echo "@set CURRENT_YEAR  $release_year" >> version.texi
echo "@set CURRENT_MONTH $release_month" >> version.texi
echo "@set PKG_NAME $PACKAGE_NAME" >> version.texi
gmake clean
gmake
rm -rf $ROOT/How-to-build-toolchain.pdf
cp How-to-build-toolchain.pdf $ROOT
popd

echo Task [III-8] /$HOST_NATIVE/pretidy/
rm -rf $INSTALLDIR_NATIVE/lib/libiberty.a
find $INSTALLDIR_NATIVE -name '*.la' -exec rm '{}' ';'

echo Task [III-9] /$HOST_NATIVE/strip_host_objects/
if [ "x$DEBUG_BUILD_OPTIONS" = "xno" ] ; then
    STRIP_BINARIES=`find $INSTALLDIR_NATIVE/bin/ -name arm-none-eabi-\*`
    for bin in $STRIP_BINARIES ; do
        strip_binary strip $bin
    done

    STRIP_BINARIES=`find $INSTALLDIR_NATIVE/arm-none-eabi/bin/ -maxdepth 1 -mindepth 1 -name \*`
    for bin in $STRIP_BINARIES ; do
        strip_binary strip $bin
    done

    STRIP_BINARIES=`find $INSTALLDIR_NATIVE/lib/gcc/arm-none-eabi/$GCC_VER/ -maxdepth 1 -name \* -perm +111 -and ! -type d`
    for bin in $STRIP_BINARIES ; do
        strip_binary strip $bin
    done
fi

echo Task [III-10] /$HOST_NATIVE/strip_target_objects/
saveenv
prepend_path PATH $INSTALLDIR_NATIVE/bin
TARGET_LIBRARIES=`find $INSTALLDIR_NATIVE/arm-none-eabi/lib -name \*.a`
for target_lib in $TARGET_LIBRARIES ; do
    arm-none-eabi-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc $target_lib || true
done

TARGET_OBJECTS=`find $INSTALLDIR_NATIVE/arm-none-eabi/lib -name \*.o`
for target_obj in $TARGET_OBJECTS ; do
    arm-none-eabi-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc $target_obj || true
done

TARGET_LIBRARIES=`find $INSTALLDIR_NATIVE/lib/gcc/arm-none-eabi/$GCC_VER -name \*.a`
for target_lib in $TARGET_LIBRARIES ; do
    arm-none-eabi-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc $target_lib || true
done

TARGET_OBJECTS=`find $INSTALLDIR_NATIVE/lib/gcc/arm-none-eabi/$GCC_VER -name \*.o`
for target_obj in $TARGET_OBJECTS ; do
    arm-none-eabi-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc $target_obj || true
done
restoreenv

echo Task [III-11] /$HOST_NATIVE/package_tbz2/
rm -f $PACKAGEDIR/$PACKAGE_NAME_NATIVE.tar.bz2
pushd $BUILDDIR_NATIVE
rm -f $INSTALL_PACKAGE_NAME
cp $ROOT/$RELEASE_FILE $INSTALLDIR_NATIVE_DOC/
cp $ROOT/$README_FILE $INSTALLDIR_NATIVE_DOC/
cp $ROOT/$LICENSE_FILE $INSTALLDIR_NATIVE_DOC/
copy_dir_clean $SRCDIR/$SAMPLES $INSTALLDIR_NATIVE/share/gcc-arm-none-eabi/$SAMPLES
ln -s $INSTALLDIR_NATIVE $INSTALL_PACKAGE_NAME
${TAR} cjf $PACKAGEDIR/$PACKAGE_NAME_NATIVE.tar.bz2 \
    --owner=0                                       \
    --group=0                                       \
    --exclude=host-$HOST_NATIVE                     \
    $INSTALL_PACKAGE_NAME/arm-none-eabi             \
    $INSTALL_PACKAGE_NAME/bin                       \
    $INSTALL_PACKAGE_NAME/lib                       \
    $INSTALL_PACKAGE_NAME/share
rm -f $INSTALL_PACKAGE_NAME
popd

echo Task [V-0] /package_sources/
pushd $PACKAGEDIR
rm -rf $PACKAGE_NAME && mkdir -p $PACKAGE_NAME/src
cp -f $SRCDIR/$ISL_PACK $PACKAGE_NAME/src/
pack_dir_clean $SRCDIR $BINUTILS $PACKAGE_NAME/src/$BINUTILS.tar.bz2
pack_dir_clean $SRCDIR $GCC $PACKAGE_NAME/src/$GCC.tar.bz2
pack_dir_clean $SRCDIR $GDB $PACKAGE_NAME/src/$GDB.tar.bz2 --exclude="gdb/testsuite/config/qemu.exp" --exclude="sim"
pack_dir_clean $SRCDIR $NEWLIB $PACKAGE_NAME/src/$NEWLIB.tar.bz2
pack_dir_clean $SRCDIR $NEWLIB_NANO $PACKAGE_NAME/src/$NEWLIB_NANO.tar.bz2
pack_dir_clean $SRCDIR $SAMPLES $PACKAGE_NAME/src/$SAMPLES.tar.bz2
pack_dir_clean $SRCDIR $BUILD_MANUAL $PACKAGE_NAME/src/$BUILD_MANUAL.tar.bz2
if [ -d $SRCDIR/$GCC_PLUGINS/ ]; then
  pack_dir_clean $SRCDIR $GCC_PLUGINS $PACKAGE_NAME/src/$GCC_PLUGINS.tar.bz2
fi

cp $ROOT/$RELEASE_FILE $PACKAGE_NAME/
cp $ROOT/$README_FILE $PACKAGE_NAME/
cp $ROOT/$LICENSE_FILE $PACKAGE_NAME/
cp $ROOT/$BUILD_MANUAL_FILE $PACKAGE_NAME/
cp $ROOT/build-*.sh $PACKAGE_NAME/
tar cjf $PACKAGE_NAME-src.tar.bz2 $PACKAGE_NAME
rm -rf $PACKAGE_NAME
popd

echo Task [V-1] /md5_checksum/
pushd $PACKAGEDIR
rm -rf md5.txt
md5 $PACKAGE_NAME_NATIVE.tar.bz2 >> md5.txt
md5 $PACKAGE_NAME-src.tar.bz2    >> md5.txt
popd

